// Instantiation of IO for vnl_diag_matrix<std::complex<float> >
#include <complex>
#include <vcl_compiler.h>
#include <vnl/io/vnl_io_diag_matrix.hxx>

VNL_IO_DIAG_MATRIX_INSTANTIATE(std::complex<float>);

