#==========================================================================
#
#   Copyright Insight Software Consortium
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0.txt
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#==========================================================================*/

#
#  Example on the use of the MedianImageFilter
#

import itk
from sys import argv

reader = itk.ImageFileReader.IUC2.New(FileName=argv[1])

# test the deduction of the template parameter from the input
filter = itk.MedianImageFilter.New(reader)
watcher = itk.XMLFilterWatcher(filter, "filter")

# test the update of the filter with the __call__ operator
# and the setting of parameter inside it
filter(Radius=eval(argv[3]))

# test the write method
itk.write(filter, argv[2])
