#==========================================================================
#
#   Copyright Insight Software Consortium
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0.txt
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#==========================================================================*/

from InsightToolkit import *
from numarray import *
from sys import argv

reader = itkImageFileReaderUC2_New()

connector = itkPyBufferUC2_New()

reader.SetFileName( argv[1] )

reader.Update()

print "ready to convert image into array"

buffer = connector.GetArrayFromImage( reader.GetOutput() )

writer = itkImageFileWriterUC2_New()

writer.SetFileName( argv[2] )

print "ready to convert array into image"

writer.SetInput( connector.GetImageFromArray( buffer ) )

writer.Update()


