// This is core/vnl/vnl_float_4.h
#ifndef vnl_float_4_h_
#define vnl_float_4_h_
#ifdef VCL_NEEDS_PRAGMA_INTERFACE
#pragma interface
#endif
//:
// \file
// \brief Contains class vnl_float_4
// \author Andrew W. Fitzgibbon, Oxford RRG
// \date   05 Aug 96
//
// \verbatim
//  Modifications
//   Peter Vanroose, 25 June 1999: vnl_vector_fixed<float,4> already instantiated
// \endverbatim

#include <vnl/vnl_T_n.h>

//: class vnl_float_4 : a vnl_vector of 4 floats.
vnl_T_n_impl(float,4);

#endif // vnl_float_4_h_
