#==========================================================================
#
#   Copyright Insight Software Consortium
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0.txt
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#==========================================================================*/

package require InsightToolkit

if {$argc == 0} {
    set fileid stdout
} else {
    set fileid [open [lindex $argv 0] w]
}

# set b [itkStringStream]
set allClasses [info command itk*]

foreach class $allClasses {
    puts $fileid "$class"
}

exit
