#==========================================================================
#
#   Copyright Insight Software Consortium
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#          http://www.apache.org/licenses/LICENSE-2.0.txt
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
#==========================================================================*/

import itk

# To verify that the lazy loading system works, we try to load
# a float image (declared in the module ITKCommon)
image=itk.Image.F3
image=itk.Image[itk.F,3]
# And a normal band image (declared in the module LevelSets)
image=itk.Image.NBNIF33
image=itk.Image[itk.NormalBandNode.IF3,3]
# The function __dir__ is overridden to load modules that
# were not loaded by the lazy loading system
itk.Image.__dir__()
