#include <vnl/xio/vnl_xio_matrix.h>
#include <vnl/xio/vnl_xio_matrix_fixed.h>
#include <vnl/xio/vnl_xio_quaternion.h>
#include <vnl/xio/vnl_xio_vector.h>
#include <vnl/xio/vnl_xio_vector_fixed.h>

int main() { return 0; }
