#include <vnl/xio/vnl_xio_matrix.hxx>
#include <vnl/xio/vnl_xio_matrix_fixed.hxx>
#include <vnl/xio/vnl_xio_quaternion.hxx>
#include <vnl/xio/vnl_xio_vector.hxx>
#include <vnl/xio/vnl_xio_vector_fixed.hxx>

int main() { return 0; }
