body {
 font-family: Inconsolata, Arial, Helvetica, sans-serif;
 padding: 0px;
 margin-top: 10px;
 margin-right: 8px;
 margin-bottom: 10px;
 margin-left: 8px;
 font-size: 12px;
 font-weight: normal;
 color: #bbbbbb;
}

sub {
 font-size: 0.83em;
}

.argument {
 background-color: #eeeeee;
 font-family: Courier;
}

.cdashfilter {
  font-family: Arial, Helvetica, sans-serif;
  font-size: 11px;
}

a:link,a:visited{
 font-size:11px;
 color:#6666ff;
 font-family: Inconsolata, Arial, Helvetica, sans-serif;
 font-weight: normal;
 text-decoration: underline;
}
a:hover{
 color:#008888;
 text-decoration: underline;
}
a:active{
 color:#6666ff;
 text-decoration: underline;
}

a.sub {
 font-size: 1em;
 text-decoration: none;
}

.toptable {
 padding: 1px;
 border: 1px solid #19262f;
 margin-top: 0px;
 margin-right: auto;
 margin-bottom: 0px;
 margin-left: auto;
}
.topline {
 background-image: url(images/bg1.gif);
 background-repeat: repeat-x;
}
.paddl {
 margin: 0px;
 padding-top: 0px;
 padding-right: 0px;
 padding-bottom: 0px;
 padding-left: 11px;
}

.topbg {
 background-image: url(images/topbg1.gif);
 background-repeat: repeat-x;
}

.topbgleft {
 background-image: url(images/top_04.jpg);
 background-repeat: repeat-x;
}

#projectlogo {
max-height : 110px;
/* IE Image max-width */
height: expression(this.height > 110 ? 110: true);
}

.paddr {
 margin: 0px;
 padding-top: 0px;
 padding-right: 11px;
 padding-bottom: 0px;
 padding-left: 0px;
}
.topdate {
 font-family: Arial, Helvetica, sans-serif;
 font-size: 14px;
 font-weight: normal;
 color: #10365b;
}
.topbg2 {
 background-image: url(images/topbg2.gif);
 background-repeat: repeat-x;
}
.footer {
 font-family: Arial, Helvetica, sans-serif;
 font-size: 14px;
 font-weight: normal;
 color: #666666;
 background-image: url(images/bgfo.gif);
 background-repeat: repeat-x;
}
.footer img{
 vertical-align: middle;
}
.normal {
 background-color : #008000;
}
.normalNew {
 background-color : #92ff89;
}
 .warning {
 background-color : #b46305;
}
 .warningNew {
 background-color : #ffbc30;
}
 .error {
 background-color : #ff6666;
}
 .errorNew {
 background-color : #ff1111;
}

 .Passed { background-color : #008000; }
 .Failed { background-color : #fdba76; }
 .Not Run { background-color : #ff6666; }
 .na { background-color : #cccccc; }
 .measurement { background-color : #b0c4de; }
 .table-heading {
 background-color : #c1c1c1;
 font-size: 13px;
}

.table-heading .nob {
 border-top-width: 0px;
 border-right-width: 0px;
 border-bottom-width: 0px;
 border-left-width: 0px;
}

.title-divider { background-color: #b0c4de; text-align: center; font-weight: bold; }
.table-heading1 {
 background-color : #3867d1;
 font-family: Arial, Helvetica, sans-serif;
 color: #FFFFFF;
}

.table-heading1 a:link{
 font-size:10px;
 color:#FFFFFF;
 font-family: Arial, Helvetica, sans-serif;
 font-weight: normal;
 text-decoration: none;
}
.table-heading1 a:visited{
 font-size:10px;
 color:#FFFFFF;
 font-family: Arial, Helvetica, sans-serif;
 font-weight: normal;
 text-decoration: none;
}

.table-heading1 h3{
 font-family: Arial, Helvetica, sans-serif;
 color: #FFFFFF;
 font-size: 17px;
 font-weight: bold;
 margin: 0px;
 padding-top: 1px;
 padding-right: 0px;
 padding-bottom: 1px;
 padding-left: 7px;
}

.tabb td, .tabb th{
 border-right-width: 2px;
 border-right-style: solid;
 border-right-color: #888888;
 padding-top: 7px;
 padding-bottom: 7px;
}
.tabb td{
 padding-top: 9px;
 padding-bottom: 9px;
}
.tabb tbody tr.odd {
  background-color:#ececec;
}
.tabb thead tr .header {
  background-repeat: no-repeat;
  background-position: bottom center;
  cursor: pointer;
}
.tabb thead tr .headerSortUp {
  background-image: url(images/asc.gif);
}
.tabb thead tr .headerSortDown {
  background-image: url(images/desc.gif);
}
.tabb thead tr .headerSortDown, .tabb thead tr .headerSortUp {
background-color: #999999;
}



 .treven {
 background-color : #ececec;
 margin: 0px;
}
 .trodd {
 background-color : #ffffff;
 margin: 0px;
 border: 0px;
}
.min {
 background-color: #fff687;
}

.tabb .nob {
 border-top-width: 0px;
 border-right-width: 0px;
 border-bottom-width: 0px;
 border-left-width: 0px;
}

.table-heading1 .nob {
 border-top-width: 0px;
 border-right-width: 0px;
 border-bottom-width: 0px;
 border-left-width: 0px;
}

.botl {
 border-bottom-width: 2px;
 border-bottom-style: solid;
 border-bottom-color: #888888;
}
#sort_11{
 border-top-width: 0px;
 border-right-width: 0px;
 border-bottom-width: 0px;
 border-left-width: 0px;
}

#nob{
 border-top-width: 0px;
 border-right-width: 0px;
 border-bottom-width: 0px;
 border-left-width: 0px;
}

.paddt {
 padding-left: 7px;
}
.total {
 background-color : #ececec;
 margin: 0px;
}
.table-heading2 {
 background-color : #156824;
 font-family: Arial, Helvetica, sans-serif;
 color: #FFFFFF;
}

.table-heading2 .nob {
 border-top-width: 0px;
 border-right-width: 0px;
 border-bottom-width: 0px;
 border-left-width: 0px;
}

.table-heading2 a:link{
 font-size:10px;
 color:#FFFFFF;
 font-family: Arial, Helvetica, sans-serif;
 font-weight: normal;
 text-decoration: none;
}
.table-heading2 a:visited{
 font-size:10px;
 color:#FFFFFF;
 font-family: Arial, Helvetica, sans-serif;
 font-weight: normal;
 text-decoration: none;
}
.table-heading2 h3{
 font-family: Arial, Helvetica, sans-serif;
 color: #FFFFFF;
 font-size: 17px;
 font-weight: bold;
 margin: 0px;
 padding-top: 1px;
 padding-right: 0px;
 padding-bottom: 1px;
 padding-left: 7px;
}
.table-heading3 a:link{
 font-size:10px;
 color:#FFFFFF;
 font-family: Arial, Helvetica, sans-serif;
 font-weight: normal;
 text-decoration: none;
}
.table-heading3 a:visited{
 font-size:10px;
 color:#FFFFFF;
 font-family: Arial, Helvetica, sans-serif;
 font-weight: normal;
 text-decoration: none;
}
.table-heading3 {
 background-color : #845924;
 font-family: Arial, Helvetica, sans-serif;
 color: #FFFFFF;
}
.table-heading3 .nob {
 border-top-width: 0px;
 border-right-width: 0px;
 border-bottom-width: 0px;
 border-left-width: 0px;
}

.table-heading3 h3{
 font-family: Arial, Helvetica, sans-serif;
 color: #FFFFFF;
 font-size: 17px;
 font-weight: bold;
 margin: 0px;
 padding-top: 1px;
 padding-right: 0px;
 padding-bottom: 1px;
 padding-left: 7px;
}.insd {
 background-image: url(images/top_02.jpg);
 margin: 0px;
 padding: 0px;
}
.inn1 {
 font-family: Arial, Helvetica, sans-serif;
 font-size: 27px;
 font-weight: bold;
 text-transform: uppercase;
 color: #fcdb34;
}
.inn2 {
 font-family: Arial, Helvetica, sans-serif;
 font-size: 24px;
 font-weight: bold;
 color: #FFFFFF;
 padding-left: 72px;
}
.insd2 {
 background-image: url(images/top_03.jpg);
 background-repeat: no-repeat;
 background-position: left top;
}
.insd3 {
 background-image: url(images/menu2_07.jpg);
 background-repeat: no-repeat;
 background-position: left top;
}
.insdd {
 background-image: url(images/top_02.jpg);
 margin: 0px;
 padding-top: 32px;
 padding-right: 0px;
 padding-bottom: 0px;
 padding-left: 143px;
}
.nav {
 list-style: none;
 margin: 0;
 padding: 0;
 font-family: Arial, Helvetica, sans-serif;
 z-index: 100;
 position: relative;

}

.nav ul {
 list-style: none;
 margin: 0;
 padding: 0;

}

.nav li {
  float: left;
  margin: 0;
  padding: 0;
  position: relative;
  /*opacity: 1.0;*/
}

.nav li a, .nav li a:link, .nav li a:active, .nav li a:visited {
 font: bold 1.00em/25px Arial, Helvetica, sans-serif;
 color: #FFFFFF;
 display: block;
 text-decoration: none;
 margin: 0px;
 padding-top: 1px;
 padding-right: 13px;
 padding-bottom: 2px;
 padding-left: 13px;
 background-image: url(images/menub1.gif);
 background-position: right bottom;
}

.nav li a:hover {
 color: #FFFFFF;
 background-image: url(images/menub2.gif);
 background-position: right bottom;
}

#nav-one li:hover a,
#nav-one li.sfHover a {
  background: #ccc;
  color: #000;
}
#nav-one li:hover ul a,
#nav-one li.sfHover ul a {
  background: #FFFFFF;
  color: #959595;
}
#nav-one li:hover ul a:hover,
#nav-one li.sfHover ul a:hover {
  background: #ccc;
  color: #000;
}


.nav li ul {
  background: #FFFFFF;
  list-style: none;
  margin: 0;
  width: 100px;
  position: absolute;
  top: -999em;
  left: -1px;
}

.nav li:hover ul,
.nav li.sfHover ul {
  top: 25px;
}

.nav ul li {
  border: 0;
  float: none;
}
.nav ul li div {

}

.nav ul li a {
 padding-right: 20px;
 width: 80px;
 white-space: nowrap;
 background-image: none;
 border-top-width: 0px;
 border-right-width: 1px;
 border-bottom-width: 1px;
 border-left-width: 1px;
 border-top-style: solid;
 border-right-style: solid;
 border-bottom-style: solid;
 border-left-style: solid;
 border-top-color: #000;
 border-right-color: #000;
 border-bottom-color: #000;
 border-left-color: #000;
}

.nav ul li a:hover {
  background: #ccc;
  color: #000;
}
.nav li a.submm{
 background-color: #3a596d;
 background-image: none;
}
#submm{
 background-color: #3a596d;
 background-image: none;
}
.bgtm {
 background-image: url(images/bgtm.gif);
 background-repeat: repeat-x;
 background-position: bottom;
}

a:hover#submm{
 background-color: #CCCCCC;
 background-image: none;
}
#activem{
 background-image: url(images/menub2.gif);
 background-position: right bottom;
}
#datepicker_div, .datepicker_inline {
 font-family: Arial, Helvetica, sans-serif;
 font-size: 14px;
 padding: 0;
 margin: 0;
 background: #ddd;
 width: 185px;
}
#datepicker_div {
 display: none;
 border: 1px solid #777;
 z-index: 10; /*must have*/
}
.datepicker_inline {
 float: left;
 display: block;
 border: 0;
}
.datepicker_dialog {
 padding: 5px !important;
 border: 4px ridge #ddd !important;
}
button.datepicker_trigger {
 width: 25px;
}
img.datepicker_trigger {
 margin: 2px;
 vertical-align: middle;
}
.datepicker_prompt {
 float: left;
 padding: 2px;
 background: #ddd;
 color: #000;
}
* html .datepicker_prompt {
 width: 185px;
}
.datepicker_control, .datepicker_links, .datepicker_header, .datepicker {
 clear: both;
 float: left;
 width: 100%;
 color: #fff;
}
.datepicker_control {
 background: #6699cc;
 padding: 2px 0px;
}
.datepicker_links {
 background: #000;
 padding: 2px 0px;
}
.datepicker_control, .datepicker_links {
 font-weight: bold;
 font-size: 80%;
 letter-spacing: 1px;
}
.datepicker_links label { /* disabled links */
 padding: 2px 5px;
 color: #888;
}
.datepicker_clear, .datepicker_prev {
 float: left;
 width: 34%;
}
.datepicker_current {
 float: left;
 width: 30%;
 text-align: center;
}
.datepicker_close, .datepicker_next {
 float: right;
 width: 34%;
 text-align: right;
}
.datepicker_header {
 padding: 1px 0 3px;
 background: #333;
 text-align: center;
 font-weight: bold;
 height: 1.3em;
}
.datepicker_header select {
 background: #333;
 color: #fff;
 border: 0px;
 font-weight: bold;
}
.datepicker {
 background: #ccc;
 text-align: center;
 font-size: 100%;
}
.datepicker a {
 display: block;
 width: 100%;
}

.datepicker_titleRow {
 background: #6699cc;
}
.datepicker_daysRow {
 background: #eee;
 color: #666;
}
.datepicker_weekCol {
 background: #777;
 color: #fff;
}
.datepicker_daysCell {
 color: #000;
 border: 1px solid #ddd;
}
.datepicker_daysCell a{
 display: block;
}
.datepicker_weekEndCell {
 background: #ddd;
}
.datepicker_daysCellOver {
 background: #fff;
 border: 1px solid #777;
}
.datepicker_unselectable {
 color: #888;
}
.datepicker_today {
 background: #fcc !important;
}
.datepicker_currentDay {
 background: #999 !important;
}

/* ________ Datepicker Links _______

** Reset link properties and then override them with !important */
#datepicker_div a, .datepicker_inline a {
 cursor: pointer;
 margin: 0;
 padding: 0;
 background: none;
 color: #000;
}
.datepicker_inline .datepicker_links a {
 padding: 0 5px !important;
}
.datepicker_control a, .datepicker_links a {
 padding: 2px 5px !important;
 color: #eee !important;
}
.datepicker_titleRow a {
 color: #eee !important;
}
.datepicker_control a:hover {
 background: #fdd !important;
 color: #333 !important;
}
.datepicker_links a:hover, .datepicker_titleRow a:hover {
 background: #ddd !important;
 color: #333 !important;
}

/* ___________ MULTIPLE MONTHS _________*/

.datepicker_multi .datepicker {
 border: 1px solid #777;
}
.datepicker_oneMonth {
 float: left;
 width: 185px;
}
.datepicker_newRow {
 clear: left;
}

/* ___________ IE6 IFRAME FIX ________ */

.datepicker_cover {
    display: none; /*sorry for IE5*/
    display/**/: block; /*sorry for IE5*/
    position: absolute; /*must have*/
    z-index: -1; /*must have*/
    /* filter: mask(); must have*/
    top: -4px; /*must have*/
    left: -4px; /*must have*/
    width: 200px; /*must have*/
    height: 200px; /*must have*/
}


.cal {
  position: absolute;
  top: -999px; /* we hide it from safari */
  left: 140px;
}


#JT_arrow_left{
  background-image: url(../images/arrow_left.gif);
  background-repeat: no-repeat;
  background-position: left top;
  position: absolute;
  z-index:101;
  left:-12px;
  height:23px;
  width:10px;
    top:-3px;
}

#JT_arrow_right{
  background-image: url(../images/arrow_right.gif);
  background-repeat: no-repeat;
  background-position: left top;
  position: absolute;
  z-index:101;
  height:23px;
  width:11px;
    top:-2px;
}

#JT {
  position: absolute;
  z-index:100;
  border: 2px solid #CCCCCC;
  background-color: #fff;
}

#JT_copy{
  padding:10px 10px 10px 10px;
  color:#333333;
}

.JT_loader{
  background-image: url(../images/loader.gif);
  background-repeat: no-repeat;
  background-position: center center;
  width:100%;
  height:12px;
}

#JT_close_left{
  background-color: #CCCCCC;
  text-align: left;
  padding-left: 8px;
  padding-bottom: 5px;
  padding-top: 2px;
  font-weight:bold;
}

#JT_close_right{
  background-color: #CCCCCC;
  text-align: left;
  padding-left: 8px;
  padding-bottom: 5px;
  padding-top: 2px;
  font-weight:bold;
}

#JT_copy p{
margin:3px 0;
}

#JT_copy img{
  padding: 1px;
  border: 1px solid #CCCCCC;
}

a.grouptrigger:link{
  font-family: Arial, Helvetica, sans-serif;
  color: #FFFFFF;
  font-size: 17px;
  font-weight: bold;
}

a.grouptrigger:visited{
  font-family: Arial, Helvetica, sans-serif;
  color: #FFFFFF;
  font-size: 17px;
  font-weight: bold;
}

/* Caution! Ensure accessibility in print and other media types... */
@media projection, screen { /* Use class for showing/hiding tab content, so that visibility can be better controlled in different media types... */
    .tabs-hide {
        display: none;
    }
}

/* Hide useless elements in print layouts... */
@media print {
    .tabs-nav {
        display: none;
    }
}

/* Skin */
.tabs-nav {
    list-style: none;
    margin: 0;
    padding: 0 0 0 4px;
}
.tabs-nav:after { /* clearing without presentational markup, IE gets extra treatment */
    display: block;
    clear: both;
    content: " ";
}
.tabs-nav li {
    float: left;
    margin: 0 0 0 1px;
    min-width: 84px; /* be nice to Opera */
}
.tabs-nav a, .tabs-nav a span {
    display: block;
    padding: 0 10px;
    background: url(images/tab.png) no-repeat;
}
.tabs-nav a {
    position: relative;
    top: 1px;
    z-index: 2;
    padding-left: 0;
    color: #000;
    font-size: 12px;
    font-weight: bold;
    line-height: 1.2;
    text-align: center;
    text-decoration: none;
    white-space: nowrap; /* required in IE 6 */
}
.tabs-nav .tabs-selected a {
    color: #000;
}

.tabs-nav .tabs-selected a, .tabs-nav a:hover, .tabs-nav a:focus, .tabs-nav a:active {
    background-position: 100% -150px;
    outline: 0; /* prevent dotted border in Firefox */
}
.tabs-nav a, .tabs-nav .tabs-disabled a:hover, .tabs-nav .tabs-disabled a:focus, .tabs-nav .tabs-disabled a:active {
    background-position: 100% -100px;
}
.tabs-nav a span {
    width: 64px; /* IE 6 treats width as min-width */
    min-width: 64px;
    height: 18px; /* IE 6 treats height as min-height */
    min-height: 18px;
    padding-top: 6px;
    padding-right: 0;
}
*>.tabs-nav a span { /* hide from IE 6 */
    width: auto;
    height: auto;
}
.tabs-nav .tabs-selected a span {
    padding-top: 7px;
}
.tabs-nav .tabs-selected a span, .tabs-nav a:hover span, .tabs-nav a:focus span, .tabs-nav a:active span {
    background-position: 0 -50px;
}
.tabs-nav a span, .tabs-nav .tabs-disabled a:hover span, .tabs-nav .tabs-disabled a:focus span, .tabs-nav .tabs-disabled a:active span {
    background-position: 0 0;
}
.tabs-nav .tabs-selected a:link, .tabs-nav .tabs-selected a:visited, .tabs-nav .tabs-disabled a:link, .tabs-nav .tabs-disabled a:visited { /* @ Opera, use pseudo classes otherwise it confuses cursor... */
    cursor: text;
}
.tabs-nav a:hover, .tabs-nav a:focus, .tabs-nav a:active { /* @ Opera, we need to be explicit again here now... */
    cursor: pointer;
}
.tabs-nav .tabs-disabled {
    opacity: .4;
}
.tabs-container {
    border-top: 1px solid #97a5b0;
    padding: 1em 8px;
    background: #fff; /* declare background color for container to avoid distorted fonts in IE while fading */
}
.tabs-loading em {
    padding: 0 0 0 20px;
    background: url(loading.gif) no-repeat 0 50%;
}
div.tab_content {
  border:1px solid #97A5B0;
}
div.tab_help{
  float:right;
  width:320px;
}
span.help_content{
  display:none;
}
#wizard{
    margin-left:auto;
    margin-right:auto;
    width:900px;
  }
/* tooltip 's css */
#tooltip{
      position:absolute;
      border:1px solid #333;
      background:#f7f5d1;
      padding:2px 5px;
      color:#333;
      display:none;
}

/* fix the position of the help tootip */

div#key {
  position: absolute;
}
pre {
  font-family: Inconsolata, Courier, Arial, Helvetica, sans-serif;
}
