/*=========================================================================
 *
 *  Copyright Insight Software Consortium
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0.txt
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *=========================================================================*/
#ifndef itkImageShapeModelEstimatorBase_hxx
#define itkImageShapeModelEstimatorBase_hxx

#include "itkImageShapeModelEstimatorBase.h"
#include "itkCommand.h"

namespace itk
{
template< typename TInputImage, typename TOutputImage >
ImageShapeModelEstimatorBase< TInputImage, TOutputImage >
::ImageShapeModelEstimatorBase(void)
{}

template< typename TInputImage, typename TOutputImage >
ImageShapeModelEstimatorBase< TInputImage, TOutputImage >
::~ImageShapeModelEstimatorBase()
{}

template< typename TInputImage, typename TOutputImage >
void
ImageShapeModelEstimatorBase< TInputImage, TOutputImage >
::GenerateData()
{
  this->EstimateShapeModels();
}

/**
 * PrintSelf
 */
template< typename TInputImage, typename TOutputImage >
void
ImageShapeModelEstimatorBase< TInputImage, TOutputImage >
::PrintSelf(std::ostream & os, Indent indent) const
{
  Superclass::PrintSelf(os, indent);

  os << indent << "                   " << std::endl;

  os << indent << "InputImage: ";
  os << m_InputImage.GetPointer() << std::endl;
} // end PrintSelf
} // namespace itk

#endif
